function [bands low_pass] = cortex_transform( img, orientations )
% Decompose image into bands using cortex transform. 
%
% [bands low_pass] = cortex_transform( img [, orientations] )
%
%  img - 2D matrix with an image to be decomposed
%  orientations - number of orientations (default = 1)
%  bands - cell array of the size 6xorientations with the band-pass and
%          orientation selective images (each one in spatial domain). The
%          last band (low-pass) is not separated into orientation-selective
%          components.
%  low_pass - cell array of the size 6 with the low-pass channels.
%          Optional.

if( ~exist( 'orientations', 'var' ) )
    orientations = 1;
end

band_count = 6;

rho = create_cycdeg_image( size(img), 1 );
theta = create_theta_image( size(img) );

mesa_filter = cortex_get_mesa_filters( rho );

for band=1:band_count-1
    dom_filter{band} = mesa_filter{band} - mesa_filter{band+1};
end

if( orientations > 1 )
    fan_filter = get_fan_filters( theta, orientations );
else
    fan_filter{1} = ones( size(img) );
end

f_img = fft2( img );

for band=1:band_count-1
    f_bp = f_img .* dom_filter{band};
    for o=1:orientations
        bands{band,o} = real( ifft2( f_bp.*fan_filter{o} ) );
    end
end
% Base band
bands{band_count,1} = real( ifft2( f_img .* mesa_filter{band_count} ) );

% Produce low-pass pyramid, if needed
if( nargout == 2 )
    low_pass{band_count} = bands{band_count,1};
    for band=band_count-1:-1:1
        low_pass{band} = low_pass{band+1};
        for o=1:orientations
            low_pass{band} = low_pass{band} + bands{band,o};
        end
    end   
end

end


function F=get_fan_filters( theta, orientations )

theta_tw = pi / orientations;

for o=1:orientations
    theta_l = (o - 1) * theta_tw - pi/2;
    
    delta_theta = abs( theta - theta_l );    
    delta_theta(delta_theta>=pi/2) = abs(pi - delta_theta(delta_theta>=pi/2));
  
    F{o} = zeros( size(theta) );
    F{o}(delta_theta<=theta_tw) = 0.5 * (1.0 + cos( pi * delta_theta(delta_theta<=theta_tw) / theta_tw));  
end

end