function [bands gain] = detection_hdrvdp( Y, name, options )
% Detection model based on the HDR-VDP
% Y - luminance map in cd/m^2
% name - string with the name of this map (shown in warnings and error
%        messages)
% options - see sources 
% bands - CSF normalized freq. bands
% gain - contrast gain sygnal (bands/gain to get 1JND normalized signal)

if( ~exist( 'options' ) )
    options = {};
end

global hdrvdp;

%Ysum = checksum( Y );
%if( isfield( hdrvdp, 'rescache' ) )
%    for k=1:length( hdrvdp.rescache )
%        if( strcmp( hdrvdp.rescache{k}.sum, Ysum ) )
%            P_vis = hdrvdp.rescache{k}.result;
%            return;
%        end
%    end
%end

view_cond.pix_per_deg = 30;
view_cond.view_dist = 0.5;
view_cond.sensitivity_correction = 1;

orient_count = 4; % the number of orientations to consider
no_c_masking = false;

% Peak contrast from Daly's CSF for L_adapt = 30 cd/m^2
daly_peak_contrast_sens = 0.006894596; 

% process options
i = 1;
while( i <= length( options ) )
    if( strcmp( options{i}, 'pix-per-deg' ) )
        view_cond.pix_per_deg = options{i+1};    
        i = i+2;
    elseif( strcmp( options{i}, 'view-dist' ) )
        view_cond.view_dist = options{i+1};
        i = i+2;
    elseif( strcmp( options{i}, 'no-contrast-masking' ) )
        no_c_masking = true;
        i = i+1;
    elseif( strcmp( options{i}, 'peak-contrast' ) )
        view_cond.sensitivity_correction = daly_peak_contrast_sens / options{i+1};
        i = i+2;
    else
        error( 'Unknown option "%s"', options{i} );
    end
end

if( any( Y(:) < 1e-4 ) )
    warning( '%s image contains values < 0.0001', name );
    Y = clamp( Y, 1e-4, 1e10 );
end

if( any( Y(:) > 1e10 ) )
    warning( '%s image contains values > 10^10', name );
    Y = clamp( Y, 1e-4, 1e10 );
end

if( any( isnan( Y(:) ) ) )
    warning( '%s image contains NaN pixel values', name );
    Y(isnan(Y)) = 1e-4;
end

% precompute common variables and structures or take them from cache

rho = create_cycdeg_image( size(Y), view_cond.pix_per_deg );
theta = create_theta_image( size(Y) );
view_cond.pupil_size = pupil_d( geomean( Y(:) ) );
otf_filter = otf_deeley( rho, view_cond.pupil_size );
csf_filters = csf_multiadapt_precompute( otf_filter );

% photoreceptor non-linearity
if( isfield( hdrvdp, 'pn_jnd' ) )
    pn_Y = hdrvdp.pn_Y;
    pn_jnd = hdrvdp.pn_jnd;
else
    [pn_Y pn_jnd] = build_jndspace();
    hdrvdp.pn_Y = pn_Y;
    hdrvdp.pn_jnd = pn_jnd;    
end
pn_jnd = pn_jnd * view_cond.sensitivity_correction;

[bands gain] = visual_pathway( Y );

% add result to cache
%if( isfield( hdrvdp, 'rescache_idx' ) )
%    rescache_idx = hdrvdp.rescache_idx;
%else
%    rescache_idx = 1;
%end
%hdrvdp.rescache{rescache_idx}.result = P_vis;
%hdrvdp.rescache{rescache_idx}.sum = Ysum;
%rescache_idx = rescache_idx+1;
%if( rescache_idx > 5 )
%    rescache_idx = 1;
%end
%hdrvdp.rescache_idx = rescache_idx;

    function [bands gain] = visual_pathway( Y )
        
        % 1 - Optical Transfer Function
        fY = fft2( Y );
        fY_post_otf = fY .* otf_filter;
        Y_post_otf = real( ifft2( fY_post_otf ) );
        
        % 2 - Photoreceptor non-linearity
        L = interp1( pn_Y, pn_jnd, clamp( Y_post_otf, pn_Y(1), pn_Y(end) ) );
        
        % 3 - CSF
        % Multi-adaptation CSF
        L_post_csf = csf_multiadapt_filter( csf_filters, L, Y_post_otf );
%        L_post_csf = L;
        

        bands = cortex_transform( L_post_csf, orient_count );
%        bands_te = cortex_transform( Y_post_otf, 6 );
%        mesa_filters = cortex_get_mesa_filters( create_cycdeg_image( size(Y), 1 ) );

%        te{length(bands),1} = ones( size(bands_te{length(bands),1} ) );

        if( no_c_masking )
            gain = {};
            return;
        end
        
        for b = 1:length(bands)
            for o=1:orient_count
                if( b == length(bands) && o > 1 )
                    break; % base band
                end            
%                y_base = Y_post_otf;
%                W = (interp1( pn_jnd, pn_Y, L + bands{b,o} ) - y_base)./y_base;
%                te{b,o} = te_daly( W, 1 );
%                bands{b,o} = bands{b,o} ./ te{b,o};

                if( true )
                    % contrast gain control with phase uncertainty
                    pu_bands{b,o} = phase_uncertainty( bands{b,o}, 180, b );
%                    y_base = Y_post_otf;
%                    if( b <= 3 )
%                        slope = 2;
%                    else
%                        slope = 1;
%                    end
                    gain{b,o} = te_simple(pu_bands{b,o}, 1 );
%                    W = (interp1( pn_jnd, pn_Y, clamp( L + pu_bands{b,o}, pn_jnd(1), pn_jnd(end) ) ) - y_base)./y_base;
 %                   gain{b,o} = te_daly( W*100, 1 );
 
%                    te{b,o} = max(ones(size(gain{b,o})), gain{b,o}.^0.5); 
%                    bands{b,o} = bands{b,o} ./ te{b,o};
                end

                if( false )
                    phase_uncetainty = 45; % in degrees
                    band_med_freq = 2.^-(b-1);
                    sigma = phase_uncetainty / 360 / band_med_freq;
                    bands{b,o} = blur_gaussian( bands{b,o}, sigma );
                end

                % Phase uncertainty - different than in the original HDRVDP
                % TODO: Filter should grow
                if( false )
                    beta = 0.2;
                    box_filter = ones( 10, 10 );
                    box_filter = box_filter / prod(size(box_filter));
                    after_pu = conv2( te{b,o}.^beta, box_filter, 'same' ).^(1/beta);
                    te{b,o} = after_pu;
                end
            end
        end
        
    end

    function filter = csf_multiadapt_precompute( otf_filter )        
        
%        filter.y_adapts = [0.01, 0.1, 1, 10, 100, 1000];       
        filter.y_adapts = [0.001 0.01 0.1 1 10 100];
        for i=1:length( filter.y_adapts )
            CSF = csf_daly( rho, theta, filter.y_adapts(i), 1, view_cond.view_dist );
            CSF_peak = max(CSF(:));
            filter.csf{i} = clamp( CSF/CSF_peak ./ otf_filter, 0, 1 ); 
        end        
    end

    function L_post_csf = csf_multiadapt_filter( filter, L, Y )
        fL = fft2( L );
        n = length( filter.y_adapts );
        for i=1:n
            fL_post_csf = fL .* filter.csf{i};
            L_csf_img = real( ifft2( fL_post_csf ) );
            L_csf_i(i,:) = L_csf_img(:);       
        end

        % interpolate
        adapt_map = log10( Y(:) );
        ind_r = interp1( log10( filter.y_adapts ), 1:n, adapt_map );
        ind_r(adapt_map<log10(filter.y_adapts(1))) = 1;
        ind_r(adapt_map>log10(filter.y_adapts(end))) = n;
        
        ind_down = floor(ind_r);
        ind_up = ceil(ind_r);
        ind_fix = (0:n:((numel(L)-1)*n))';
        ind_up = ind_up + ind_fix;
        ind_down = ind_down + ind_fix;
        ratio = mod( ind_r, 1 );
        
        L_post_csf = zeros( size(L) );
        L_post_csf(:) = L_csf_i(ind_up).*ratio + L_csf_i(ind_down).*(1-ratio);        
    end

    function Y = phase_uncertainty( X, phase, band )
        
        band_med_freq = 2.^-(band-1);
        sigma = phase / 360 / band_med_freq;
        
        ksize2 = round(sigma*1.8);
        gauss_1d = exp( -(-ksize2:ksize2).^2/(2*sigma^2) );
        gauss_1d = gauss_1d / sum( gauss_1d );
        beta = 20;
        
        Y = conv2( abs(X).^beta, gauss_1d, 'same' );
        Y = conv2( Y, gauss_1d', 'same' ).^(1/beta);
        
    end

    function gain = te_simple( mask, slope )
        m = abs( mask );
        gain = ones( size(mask) );
        gain(m>1) = m(m>1).^slope;
    end

end